/*
 * httpfs.c -- HTTP access virtual filesystem for Linux
 * Copyright (C) 2001  Andrew Main
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * main todo list
 *
 * + accept literal IPv4 addresses as well as host names
 *
 * + be a proper HTTP cache -- follow HTTP caching rules
 *
 * + do a HEAD request earlier, show metadata in inode
 *
 * + find a neater way to handle non-existent resources
 *
 * + make response parsing less offensively char-bashing
 *
 * + revise the HTTP dialogue to make sure it matches the standard properly
 *
 * + do something about the potentially-unbounded memory requirements
 *
 * + make data objects regular files instead of directories
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/stddef.h>
#include <linux/net.h>
#include <linux/in.h>
#include <net/sock.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/byteorder.h>
#include "resolv.h"
#include "resolv_numbers.h"
#include "resolv_dn_util.h"

static int xdigitval(char c)
{
	switch(c) {
		case '0': return 0;
		case '1': return 1;
		case '2': return 2;
		case '3': return 3;
		case '4': return 4;
		case '5': return 5;
		case '6': return 6;
		case '7': return 7;
		case '8': return 8;
		case '9': return 9;
		case 'A': case 'a': return 10;
		case 'B': case 'b': return 11;
		case 'C': case 'c': return 12;
		case 'D': case 'd': return 13;
		case 'E': case 'e': return 14;
		case 'F': case 'f': return 15;
		default: return -1;
	}
}

static void minescape_url(char **out_p, char const **in_p,
	char const *unescaped, char const *reserved)
{
	static char const * const always_unescaped =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
		"0123456789" "-_.!~*'()";
	char *out = *out_p;
	char const *in = *in_p;
	for(;;) {
		char c = *in;
		if(c && (strchr(unescaped, c) || strchr(always_unescaped, c) ||
				strchr(reserved, c))) {
			in++;
			*out++ = c;
		} else if(c == '%') {
			int v1, v2;
			v1 = xdigitval(in[1]);
			if(v1 == -1)
				break;
			v2 = xdigitval(in[2]);
			if(v2 == -1)
				break;
			c = (v1<<4) + v2;
			if(c && (strchr(unescaped, c) ||
					strchr(always_unescaped, c))) {
				in += 3;
				*out++ = c;
			} else {
				out[0] = in[0];
				out[1] = in[1];
				out[2] = in[2];
				in += 3;
				out += 3;
			}
		} else
			break;
	}
	*out_p = out;
	*in_p = in;
}

#define MY_SUPER_DATA(S) ((struct my_super_data *)((S)->u.generic_sbp))

struct my_super_data {
	spinlock_t nextinode_lock;
	ino_t nextinode;
};

struct my_host_inode_data;

#define MY_PATH_INODE_DATA(I) ((struct my_path_inode_data *)((I)->u.generic_ip))

struct my_path_inode_data {
	struct my_host_inode_data *host;
	size_t pathlen;
	char *path;   /* as it appears in the URL, e.g., "/kumquats" */
};

#define MY_HOST_INODE_DATA(I) ((struct my_host_inode_data *)((I)->u.generic_ip))

struct my_host_inode_data {
	struct my_path_inode_data pd;
	domainname_ptr_t hostname_domain;
	u16 port;
	size_t hostlen;
	char *host;   /* as it appears in the URL, e.g., "www.foo.org:88" */
};

#define MY_PATH_FILE_DATA(F) ((struct my_path_file_data *)((F)->private_data))

struct my_path_file_data {
	size_t content_length;
	char *content;
};

static struct socket *get_tcp_conn(struct my_host_inode_data *hd)
{
	int result;
	struct sockaddr_in addr;
	struct rrset *nsaddrs;
	struct rdata *rd;
	int rc;
	struct socket *sock;
	nsaddrs = resolv_getrrset(hd->hostname_domain, C_IN, T_A);
	if(IS_ERR(nsaddrs)) {
		result = PTR_ERR(nsaddrs);
		goto err_dnsfail;
	}
	if(!(rc = nsaddrs->rrcount)) {
		result = -EHOSTUNREACH;
		goto err_noaddrs;
	}
	result = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
	if(result < 0)
		goto err_nosocket;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(hd->port);
	for(rd = &nsaddrs->rdata[0]; rc--; ) {
		u16 rdlen = get_unaligned(&rd->length);
		if(rdlen == 4) {
			memcpy(&addr.sin_addr, rd->data, 4);
			result = sock->ops->connect(sock,
					(struct sockaddr *)&addr, sizeof(addr),
					O_RDWR);
			if(!result) {
				kfree(nsaddrs);
				return sock;
			}
		}
		rd = (struct rdata *)(rd->data + rdlen);
	}
	sock_release(sock);
err_nosocket:
err_noaddrs:
	kfree(nsaddrs);
err_dnsfail:
	return ERR_PTR(result);
}

static int http_get_header(struct socket *sock, struct my_path_file_data *pfd)
{
	mm_segment_t oldfs;
	struct msghdr msgh;
	ssize_t result;
	struct iovec iov;
	size_t buflen = 512, oldbuflen = 3;
	char *buffer, *p;
	msgh.msg_name = NULL;
	msgh.msg_namelen = 0;
	msgh.msg_iov = &iov;
	msgh.msg_iovlen = 1;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_flags = 0;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	for(;;) {
		buffer = kmalloc(buflen, GFP_KERNEL);
		if(!buffer)
			goto err_enomem;
		iov.iov_base = buffer;
		iov.iov_len = buflen-1;
		result = sock_recvmsg(sock, &msgh, buflen-1,
			MSG_PEEK|MSG_WAITALL);
		if(result < 0)
			goto err_free;
		buffer[result] = 0;
		p = strstr(buffer+oldbuflen-3, "\r\n\r\n");
		if(p)
			break;
		if(result != buflen-1)
			goto err_eio;
		kfree(buffer);
		oldbuflen = buflen;
		buflen <<= 1;
	}
	iov.iov_len = p+4-buffer;
	result = sock_recvmsg(sock, &msgh, p+4-buffer, MSG_WAITALL);
	if(result < 0)
		goto err_free;
	if(result >= 0 && result != p+4-buffer)
		goto err_eio;
	p[2] = 0;
	if(memcmp(buffer, "HTTP/", 5))
		goto err_eio;
	p = buffer+5;
	if(!(*p >= '0' && *p <= '9'))
		goto err_eio;
	while(*++p >= '0' && *p <= '9')
		;
	if(*p++ != '.')
		goto err_eio;
	if(!(*p >= '0' && *p <= '9'))
		goto err_eio;
	while(*++p >= '0' && *p <= '9')
		;
	if(*p != ' ' || !(p[1] >= '0' && p[1] <= '9') ||
			!(p[2] >= '0' && p[2] <= '9') ||
			!(p[3] >= '0' && p[3] <= '9') || p[4] != ' ')
		goto err_eio;
	if(p[1] == '1') {
		result = -EAGAIN;
		goto err_free;
	}
	for(p = buffer; *p; p++)
		if(*p >= 'A' && *p <= 'Z')
			*p += 'a'-'A';
	p = strstr(buffer, "\r\ntransfer-encoding:");
	if(p) {
		p += 20;
		while(*p == ' ' || *p == '\t')
			p++;
		if(memcmp(p, "identity\r\n", 10)) {
			pfd->content = ERR_PTR(-1);
			result = 0;
			goto err_free;
		}
	}
	p = strstr(buffer, "\r\ncontent-length:");
	if(p) {
		size_t sz = 0;
		p += 17;
		while(*p == ' ' || *p == '\t')
			p++;
		while(*p >= '0' && *p <= '9')
			sz = sz*10 + *p++-'0';
		if(*p == '\r') {
			pfd->content_length = sz;
			pfd->content = ERR_PTR(-2);
			result = 0;
			goto err_free;
		}
	}
	pfd->content = ERR_PTR(-3);
	result = 0;
err_free:
	kfree(buffer);
err:
	set_fs(oldfs);
	return result;
err_eio:
	result = -EIO;
	goto err_free;
err_enomem:
	result = -EIO;
	goto err;
}

static int http_get_body(struct socket *sock, struct my_path_file_data *pfd)
{
	mm_segment_t oldfs;
	struct msghdr msgh;
	ssize_t result;
	struct iovec iov;
	struct list_head chunklist;
	size_t totallen = 0;
	struct chunk {
		struct list_head lh;
		size_t length;
		char data[0];
	};
	oldfs = get_fs(); set_fs(KERNEL_DS);
	msgh.msg_name = NULL;
	msgh.msg_namelen = 0;
	msgh.msg_iov = &iov;
	msgh.msg_iovlen = 1;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_flags = 0;
	INIT_LIST_HEAD(&chunklist);
	switch(PTR_ERR(pfd->content)) {
		case -1: {
			/* chunked */
			for(;;) {
				struct chunk *ch;
				size_t chsize = 0;
				char c;
				for(;;) {
					iov.iov_base = &c;
					iov.iov_len = 1;
					result = sock_recvmsg(sock, &msgh, 1,
							MSG_WAITALL);
					if(result < 0)
						goto err;
					if(result != 1)
						goto err_eio;
					if(xdigitval(c) == -1)
						break;
					chsize = (chsize<<4) | xdigitval(c);
				}
				if(c == ' ') {
					do {
						iov.iov_base = &c;
						iov.iov_len = 1;
						result = sock_recvmsg(sock,
							&msgh, 1, MSG_WAITALL);
						if(result < 0)
							goto err;
						if(result != 1)
							goto err_eio;
					} while(c == ' ');
				} else if(c == ';') {
					do {
						iov.iov_base = &c;
						iov.iov_len = 1;
						result = sock_recvmsg(sock,
							&msgh, 1, MSG_WAITALL);
						if(result < 0)
							goto err;
						if(result != 1)
							goto err_eio;
					} while(c != '\r');
				} else if(c != '\r')
					goto err_eio;
				iov.iov_base = &c;
				iov.iov_len = 1;
				result = sock_recvmsg(sock, &msgh, 1,
						MSG_WAITALL);
				if(result < 0)
					goto err;
				if(result != 1 || c != '\n')
					goto err_eio;
				if(!chsize)
					break;
				ch = kmalloc(offsetof(struct chunk, data) +
							chsize, GFP_KERNEL);
				if(!ch)
					goto err_enomem;
				list_add_tail(&ch->lh, &chunklist);
				iov.iov_base = ch->data;
				iov.iov_len = chsize;
				result = sock_recvmsg(sock, &msgh, chsize,
						MSG_WAITALL);
				if(result < 0)
					goto err;
				if(result != chsize)
					goto err_eio;
				ch->length = result;
				totallen += result;
				{
					char crlf[2];
					iov.iov_base = crlf;
					iov.iov_len = 2;
					result = sock_recvmsg(sock, &msgh, 2,
							MSG_WAITALL);
					if(result < 0)
						goto err;
					if(result != 2 || crlf[0] != '\r' ||
							crlf[1] != '\n')
						goto err_eio;
				}
			}
		} break;
		case -2: {
			/* known length */
			pfd->content = kmalloc(pfd->content_length, GFP_KERNEL);
			if(!pfd->content)
				goto err_enomem;
			iov.iov_base = pfd->content;
			iov.iov_len = pfd->content_length;
			result = sock_recvmsg(sock, &msgh, pfd->content_length,
					MSG_WAITALL);
			if(result != pfd->content_length) {
				kfree(pfd->content);
				goto err_eio;
			}
			set_fs(oldfs);
			return 0;
		} break;
		case -3: {
			/* unknown length */
			size_t chsize = 1024;
			do {
				struct chunk *ch =
					kmalloc(offsetof(struct chunk, data) +
							chsize, GFP_KERNEL);
				if(!ch)
					goto err_enomem;
				list_add_tail(&ch->lh, &chunklist);
				iov.iov_base = ch->data;
				iov.iov_len = chsize;
				result = sock_recvmsg(sock, &msgh, chsize,
						MSG_WAITALL);
				if(result < 0)
					goto err;
				ch->length = result;
				totallen += result;
			} while(result == chsize);
		} break;
		default: goto err_eio;
	}
	{
		struct list_head *lp;
		char *p = kmalloc(totallen, GFP_KERNEL);
		if(!p)
			goto err_enomem;
		pfd->content_length = totallen;
		pfd->content = p;
		for(lp = chunklist.next; lp != &chunklist; lp = lp->next) {
			struct chunk *ch = list_entry(lp, struct chunk, lh);
			memcpy(p, ch->data, ch->length);
			p += ch->length;
		}
	}
	result = 0;
err:
	set_fs(oldfs);
	{
		struct list_head *lp;
		for(lp = chunklist.next; lp != &chunklist; ) {
			struct list_head *nlp = lp->next;
			kfree(list_entry(lp, struct chunk, lh));
			lp = nlp;
		}
	}
	return result;
err_enomem:
	result = -ENOMEM;
	goto err;
err_eio:
	result = -EIO;
	goto err;
}

static int send_to_socket(struct socket *sock, size_t niov, struct iovec *iov)
{
	mm_segment_t oldfs;
	struct msghdr msgh;
	size_t length;
	ssize_t result;
	msgh.msg_name = NULL;
	msgh.msg_namelen = 0;
	msgh.msg_iov = iov;
	msgh.msg_iovlen = niov;
	msgh.msg_control = NULL;
	msgh.msg_controllen = 0;
	msgh.msg_flags = 0;
	for(length = 0; niov--; )
		length += iov++->iov_len;
	oldfs = get_fs(); set_fs(KERNEL_DS);
	for(;;) {
		result = sock_sendmsg(sock, &msgh, length);
		if(result < 0 || result == length)
			break;
		while(msgh.msg_iov->iov_len <= result) {
			result -= msgh.msg_iov->iov_len;
			length -= msgh.msg_iov->iov_len;
			msgh.msg_iov++;
			msgh.msg_iovlen--;
		}
		msgh.msg_iov->iov_base = ((u8 *)msgh.msg_iov->iov_base)+result;
		msgh.msg_iov->iov_len -= result;
		length -= result;
	}
	set_fs(oldfs);
	/* TODO: ensure TCP push */
	return result < 0 ? result : 0;
}

static int http_send_query(struct socket *sock, struct my_path_inode_data *pd)
{
	struct iovec iov[5];
	iov[0].iov_base = "GET ";
	iov[0].iov_len = 4;
	iov[1].iov_base = pd->path;
	iov[1].iov_len = pd->pathlen;
	iov[2].iov_base = " HTTP/1.1\r\nConnection: close\r\nHost: ";
	iov[2].iov_len = 36;
	iov[3].iov_base = pd->host->host;
	iov[3].iov_len = pd->host->hostlen;
	iov[4].iov_base = "\r\n\r\n";
	iov[4].iov_len = 4;
	return send_to_socket(sock, 5, iov);
}

static int http_read(struct my_path_file_data *pfd,
	struct my_path_inode_data *pd)
{
	struct socket *server;
	int result;
	server = get_tcp_conn(pd->host);
	if(IS_ERR(server)) {
		result = PTR_ERR(server);
		goto err_noconn;
	}
	result = http_send_query(server, pd);
	if(result)
		goto err_send_failed;
	do {
		result = http_get_header(server, pfd);
	} while(result == EAGAIN);
	if(result)
		goto err_bad_header;
	result = http_get_body(server, pfd);
err_bad_header:
err_send_failed:
	sock_release(server);
err_noconn:
	return result;
}

static struct file_operations root_file_ops = {
	read: generic_read_dir,
};

static ssize_t my_path_read(struct file *f, char *buf, size_t count,
	loff_t *ppos)
{
	struct my_path_file_data *pfd = MY_PATH_FILE_DATA(f);
	if(*ppos >= pfd->content_length)
		return 0;
	if(count > pfd->content_length - *ppos)
		count = pfd->content_length - *ppos;
	copy_to_user(buf, pfd->content + *ppos, count);
	*ppos += count;
	return count;
}

static int my_path_open(struct inode *i, struct file *f)
{
	struct my_path_file_data *pfd;
	int result;
	if(!MY_PATH_INODE_DATA(i)->pathlen)
		return -ENOENT;
	pfd = kmalloc(sizeof(*pfd), GFP_KERNEL);
	if(!pfd)
		goto err_enomem;
	result = http_read(pfd, MY_PATH_INODE_DATA(i));
	if(result)
		goto err_read_failed;
	MY_PATH_FILE_DATA(f) = pfd;
	return 0;
err_enomem:
	return -ENOMEM;
err_read_failed:
	kfree(pfd);
	return result;
}

static int my_path_release(struct inode *i, struct file *f)
{
	struct my_path_file_data *pfd = MY_PATH_FILE_DATA(f);
	kfree(pfd->content);
	kfree(pfd);
	return 0;
}

static struct file_operations my_path_file_ops = {
	open: my_path_open,
	release: my_path_release,
	read: my_path_read,
};

static struct inode_operations my_path_inode_ops;

static struct inode *mk_path_inode(struct super_block *s,
	struct my_path_inode_data *pd)
{
	struct inode *i;
	ino_t inum;
	spin_lock(&MY_SUPER_DATA(s)->nextinode_lock);
	inum = MY_SUPER_DATA(s)->nextinode++;
	spin_unlock(&MY_SUPER_DATA(s)->nextinode_lock);
	i = iget(s, inum);
	if(!i)
		goto fail_noinode;
	i->i_op = &my_path_inode_ops;
	i->i_fop = &my_path_file_ops;
	i->i_mode = S_IFDIR|0555;
	MY_PATH_INODE_DATA(i) = pd;
	return i;
fail_noinode:
	kfree(pd);
	return ERR_PTR(-ENOMEM);
}

static struct my_path_inode_data *interpret_pathseg(
	struct my_path_inode_data *parentpd, struct qstr *name)
{
	struct my_path_inode_data *pd =
		kmalloc(sizeof(*pd) + parentpd->pathlen + 1 + name->len + 1,
			GFP_KERNEL);
	char *p;
	char const *np;
	if(!pd)
		return ERR_PTR(-ENOMEM);
	pd->host = parentpd->host;
	pd->path = ((char *)pd) + sizeof(*pd);
	memcpy(pd->path, parentpd->path, parentpd->pathlen);
	p = pd->path + parentpd->pathlen;
	*p++ = '/';
	np = name->name;
	if(name->len != 1 || *np != '%') {
		minescape_url(&p, &np, "", ";/?:@&=+$,");
		if(*np)
			goto err_enoent;
	}
	*p = 0;
	pd->pathlen = p - pd->path;
	return pd;
err_enoent:
	kfree(pd);
	return ERR_PTR(-ENOENT);
}

static struct dentry *my_path_lookup(struct inode *parenti,
	struct dentry *childd)
{
	struct my_path_inode_data *pd =
		interpret_pathseg(MY_PATH_INODE_DATA(parenti), &childd->d_name);
	struct inode *i;
	int result;
	if(IS_ERR(pd)) {
		result = PTR_ERR(pd);
		goto err;
	}
	i = mk_path_inode(parenti->i_sb, pd);
	if(IS_ERR(i))
		return ERR_PTR(PTR_ERR(i));
	d_add(childd, i);
	return NULL;
err:
	if(result != -ENOENT)
		return ERR_PTR(result);
	d_add(childd, NULL);
	return NULL;
}

static struct inode_operations my_path_inode_ops = {
	lookup: my_path_lookup,
};

static struct inode *mk_host_inode(struct super_block *s,
	struct my_host_inode_data *hd)
{
	hd->pd.pathlen = 0;
	hd->pd.path = hd->host + hd->hostlen;
	hd->pd.host = hd;
	return mk_path_inode(s, &hd->pd);
}

static int host_status_cb(void *magic, struct rrset const *rrs)
{
	return 0;
}

static int host_status(domainname_ptr_t domainname)
{
	return resolv_userrset(domainname, C_IN, T_A, host_status_cb, NULL);
}

static int text2domainname(domainname_ptr_t out, char const *p)
{
	size_t plen = strlen(p);
	return -1 == domainname_text_to_bin(out, plen+2, p, plen) ? -ENOENT : 0;
}

static struct my_host_inode_data *interpret_hostname(struct qstr *name)
{
	struct my_host_inode_data *hd =
		kmalloc(sizeof(*hd) + name->len*2 + 3, GFP_KERNEL);
	char *p, *q;
	char const *np;
	char nc;
	int result;
	if(!hd)
		return ERR_PTR(-ENOMEM);
	p = ((char *)hd) + sizeof(*hd);
	hd->host = p;
	np = name->name;
	minescape_url(&p, &np, "&=+$,", "");
	nc = *np++;
	if(nc && nc != ':')
		goto err_enoent;
	*p++ = 0;
	if(nc && *np) {
		char *pp = p;
		int portno = 0;
		minescape_url(&pp, &np, "&=+$,", "");
		if(*np)
			goto err_enoent;
		*pp = 0;
		for(pp = p; *pp; pp++) {
			if(*pp < '0' || *pp > '9')
				goto err_enoent;
			portno = portno*10 + *pp-'0';
			if(portno > 65535)
				goto err_enoent;
		}
		if(!portno)
			goto err_enoent;
		hd->port = portno;
	} else
		hd->port = 80;
	q = p;
	if(hd->port != 80)
		q += sprintf(q, "%u", (unsigned)hd->port) + 1;
	hd->hostlen = q-1 - hd->host;
	hd->hostname_domain = (domainname_ptr_t)q;
	result = text2domainname(hd->hostname_domain, hd->host);
	if(result)
		goto err;
	if(p != q)
		p[-1] = ':';
	return hd;
err_enoent:
	result = -ENOENT;
err:
	kfree(hd);
	return ERR_PTR(result);
}

static struct dentry *my_root_lookup(struct inode *parenti,
	struct dentry *childd)
{
	struct my_host_inode_data *hd = interpret_hostname(&childd->d_name);
	struct inode *i;
	int result;
	if(IS_ERR(hd)) {
		result = PTR_ERR(hd);
		goto err;
	}
	result = host_status(hd->hostname_domain);
	if(result)
		goto err_free;
	i = mk_host_inode(parenti->i_sb, hd);
	if(IS_ERR(i))
		return ERR_PTR(PTR_ERR(i));
	d_add(childd, i);
	return NULL;
err_free:
	kfree(hd);
err:
	if(result != -ENOENT)
		return ERR_PTR(result);
	d_add(childd, NULL);
	return NULL;
}

static struct inode_operations my_root_inode_ops = {
	lookup: my_root_lookup,
};

static void my_read_inode(struct inode *i)
{
	i->i_uid = 0;
	i->i_gid = 0;
	i->i_nlink = 1;
	i->i_size = 0;
	i->i_mtime = i->i_atime = i->i_ctime = CURRENT_TIME;
	i->i_blocks = i->i_blksize = 0;
}

static void my_put_inode(struct inode *i)
{
	if(i->i_op == &my_path_inode_ops) {
		struct my_path_inode_data *pd = MY_PATH_INODE_DATA(i);
		kfree(pd);
	}
}

static int my_statfs(struct super_block *sb, struct statfs *buf)
{
	buf->f_type = 0;
	buf->f_bsize = 1;
	buf->f_bfree = 0;
	buf->f_bavail = 0;
	buf->f_ffree = 0;
	buf->f_namelen = 0;
	return 0;
}

static void my_put_super(struct super_block *s)
{
	kfree(MY_SUPER_DATA(s));
}

static struct super_operations my_super_ops = {
	put_super: my_put_super,
	statfs: my_statfs,
	read_inode: my_read_inode,
	put_inode: my_put_inode,
};

static inline struct inode *mk_root_inode(struct super_block *s)
{
	struct inode *i = iget(s, 1);
	if(!i)
		return NULL;
	i->i_op = &my_root_inode_ops;
	i->i_fop = &root_file_ops;
	i->i_mode = S_IFDIR|0111;
	return i;
}

static struct super_block *my_read_super(struct super_block *s, void *data,
	int silent)
{
	struct inode *root_inode;
	s->s_blocksize = 1;
	s->s_blocksize_bits = 0;
	s->s_op = &my_super_ops;
	s->s_magic = 0;
	MY_SUPER_DATA(s) = kmalloc(sizeof(struct my_super_data), GFP_KERNEL);
	if(!MY_SUPER_DATA(s))
		goto fail_nomydata;
	spin_lock_init(&MY_SUPER_DATA(s)->nextinode_lock);
	MY_SUPER_DATA(s)->nextinode = 2;
	root_inode = mk_root_inode(s);
	if(!root_inode)
		goto fail_norootinode;
	s->s_root = d_alloc_root(root_inode);
	if(!s->s_root)
		goto fail_norootdentry;
	return s;
fail_norootdentry:
	iput(root_inode);
fail_norootinode:
	kfree(MY_SUPER_DATA(s));
fail_nomydata:
	return NULL;
}

static DECLARE_FSTYPE(my_fs_type, "httpfs", my_read_super, 0);

static int __init my_init(void)
{
	return register_filesystem(&my_fs_type);
}

static void __exit my_exit(void)
{
	unregister_filesystem(&my_fs_type);
}

module_init(my_init)
module_exit(my_exit)
EXPORT_NO_SYMBOLS;

#ifdef MODULE_AUTHOR
MODULE_AUTHOR("Andrew Main (Zefram) <zefram@fysh.org>");
#endif /* MODULE_AUTHOR */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif /* MODULE_LICENSE */
#ifdef MODULE_DESCRIPTION
MODULE_DESCRIPTION("HTTP access virtual filesystem");
#endif /* MODULE_DESCRIPTION */
